local _G = getfenv(0)
local LibStub = _G.LibStub

local candy = LibStub("LibCandyBar-3.0")
local media = LibStub("LibSharedMedia-3.0")

local major = "CooldownFrames-1.0"
local minor = tonumber(("$Rev: 1 $"):match("(%d+)")) or 1

local lib = LibStub:NewLibrary(major, minor)
if not lib then return end

local CreateFrame = CreateFrame
local GameFontNormal = GameFontNormal
local pairs = pairs
local unpack = unpack
local wipe = wipe
local tsort = table.sort

local defaulttexture = "Interface\\TargetingFrame\\UI-StatusBar"

--list of specs
local specs = {
	["default"] = {
		name = "Unknown",
		color = {r=1.0, g=0.0, b=0.0},
		icon = "Interface\\Icons\\inv_misc_questionmark",
	},
	["DEATHKNIGHT"] = {
		[1] = {
			name = "Blood",
			color = {r=1.0, g=0.0, b=0.0},
			icon = "Interface\\Icons\\spell_deathknight_bloodpresence",
		},
		[2] = {
			name = "Frost",
			color = {r=0.3, g=0.5, b=1.0},
			icon = "Interface\\Icons\\spell_deathknight_frostpresence",
		},
		[3] = {
			name = "Unholy",
			color = {r=0.2, g=0.8, b=0.2},
			icon = "Interface\\Icons\\spell_deathknight_unholypresence",
		}
	},
	 
	["DRUID"] = {
		[1] = {
			name = "Balance",
			color = {r=0.8, g=0.3, b=0.8},
			icon = "Interface\\Icons\\spell_nature_starfall",
		},
		[2] = {
			name = "Feral",
			color = {r=1.0, g=0.0, b=0.0},
			icon = "Interface\\Icons\\ability_racial_bearform",
		},
		[3] = {
			name = "Restoration",
			color = {r=0.4, g=0.8, b=0.2},
			icon = "Interface\\Icons\\spell_nature_healingtouch",
		}   
	},
	 
	["HUNTER"] = {
		[1] = {
			name = "Beast Mastery",
			color = {r=1.0, g=0.0, b=0.3},
			icon = "Interface\\Icons\\ability_hunter_bestialdiscipline",
		},
		[2] = {
			name = "Marksmanship",
			color = {r=0.3, g=0.6, b=1.0},
			icon = "Interface\\Icons\\ability_hunter_focusedaim",
		},
		[3] = {
			name = "Survival",
			color = {r=1.0, g=0.6, b=0.0},
			icon = "Interface\\Icons\\ability_hunter_camouflage",
		}
	},
	 
	["MAGE"] = {
		[1] = {
			name = "Arcane",
			color = {r=0.7, g=0.2, b=1.0},
			icon = "Interface\\Icons\\spell_holy_magicalsentry",
		},
		[2] = {
			name = "Fire",
			color = {r=1.0, g=0.5, b=0.0},
			icon = "Interface\\Icons\\spell_fire_firebolt02",
		},
		[3] = {
			name = "Frost",
			color = {r=0.3, g=0.6, b=1.0},
			icon = "Interface\\Icons\\spell_frost_frostbolt02",
		}
	},
	 
	["PALADIN"] = {
		[1] = {
			name = "Holy",
			color = {r=1.0, g=0.5, b=0.0},
			icon = "Interface\\Icons\\spell_holy_holybolt",
		},
		[2] = {
			name = "Protection",
			color = {r=0.3, g=0.5, b=1.0},
			icon = "Interface\\Icons\\ability_paladin_shieldofthetemplar",
		},
		[3] = {
			name = "Retribution",
			color = {r=1.0, g=0.0, b=0.0},
			icon = "Interface\\Icons\\spell_holy_auraoflight",
		}
	},
	 
	["PRIEST"] = {
		[1] = {
			name = "Discipline",
			color = {r=1.0, g=0.5, b=0.0},
			icon = "Interface\\Icons\\spell_holy_powerwordshield",
		},
		[2] = {
			name = "Holy",
			color = {r=0.6, g=0.6, b=1.0},
			icon = "Interface\\Icons\\spell_holy_guardianspirit",
		},
		[3] = {
			name = "Shadow",
			color = {r=0.7, g=0.4, b=0.8},
			icon = "Interface\\Icons\\spell_shadow_shadowwordpain",
		}
	},
	 
	["ROGUE"] = {
		[1] = {
			name = "Assassination",
			color = {r=0.5, g=0.8, b=0.5},
			icon = "Interface\\Icons\\ability_rogue_eviscerate",
		},
		[2] = {
			name = "Combat",
			color = {r=1.0, g=0.5, b=0.0},
			icon = "Interface\\Icons\\ability_backstab",
		},
		[3] = {
			name = "Subtlety",
			color = {r=0.3, g=0.5, b=1.0},
			icon = "Interface\\Icons\\ability_stealth",
		}
	},
	 
	["SHAMAN"] = {
		[1] = {
			name = "Elemental",
			color = {r=0.8, g=0.2, b=0.8},
			icon = "Interface\\Icons\\spell_nature_lightning",
		},
		[2] = {
			name = "Enhancement",
			color = {r=0.3, g=0.5, b=1.0},
			icon = "Interface\\Icons\\spell_nature_lightningshield",
		},
		[3] = {
			name = "Restoration",
			color = {r=0.2, g=0.8, b=0.4},
			icon = "Interface\\Icons\\spell_nature_magicimmunity",
		}
	},
	 
	["WARLOCK"] = {
		[1] = {
			name = "Affliction",
			color = {r=0.0, g=1.0, b=0.6},
			icon = "Interface\\Icons\\spell_shadow_deathcoil",
		},
		[2] = {
			name = "Demonology",
			color = {r=1.0, g=0.0, b=0.0},
			icon = "Interface\\Icons\\spell_shadow_metamorphosis",
		},
		[3] = {
			name = "Destruction",
			color = {r=1.0, g=0.5, b=0.0},
			icon = "Interface\\Icons\\spell_shadow_rainoffire",
		}
	},
	 
	["WARRIOR"] = {
		[1] = {
			name = "Arms",
			color = {r=1.0, g=0.72, b=0.1},
			icon = "Interface\\Icons\\ability_warrior_savageblow",
		},
		[2] = {
			name = "Fury",
			color = {r=1.0, g=0.0, b=0.0},
			icon = "Interface\\Icons\\ability_warrior_innerrage",
		},
		[3] = {
			name = "Protection",
			color = {r=0.3, g=0.5, b=1.0},
			icon = "Interface\\Icons\\ability_warrior_defensivestance",
		}
	},
}

local defaulticon = specs["default"].icon

function lib:GetSpecs()
	return specs
end

local function barstopped(callback, bar)
	local anchor = bar:Get("anchor")
	local spellID = bar:Get("spellID")

	if anchor.running[bar] then
		anchor.running[bar] = nil
	end
end

local function sortBars(a, b)
	return (a.remaining > b.remaining)
end

local function rearrangeBars(anchor)
	local tmp = {}
	for bar in pairs(anchor.running) do
		tmp[#tmp+1] = bar
	end
	tsort(tmp, sortBars)
	local lastBar = nil
	local rowFirst = nil
	local barNo = 0
	local width = anchor:GetWidth()
	for i, bar in ipairs(tmp) do
		bar:ClearAllPoints()
		bar:SetWidth(width/anchor.columns)
		if anchor.growup then
			if (barNo % anchor.columns) == 0 then
				bar:SetPoint("BOTTOMLEFT", rowFirst or anchor, "TOPLEFT")
				rowFirst = bar
			else
				bar:SetPoint("LEFT", lastBar or anchor, "RIGHT")
			end
		else
			if (barNo % anchor.columns) == 0 then
				bar:SetPoint("TOPLEFT", rowFirst or anchor, "BOTTOMLEFT")
				rowFirst = bar
			else
				bar:SetPoint("LEFT", lastBar or anchor, "RIGHT")
			end
		end
		lastBar = bar
		barNo = barNo + 1
	end
--	wipe(tmp)
end

local function getBar(anchor, id)
	local bar = false
	for k in pairs(anchor.running) do
		if k:Get("id") == id then
			bar = true
			break
		end
	end
	return bar
end

local function onDragHandleMouseDown(self)
	self:GetParent():StartSizing("BOTTOMRIGHT")
end

local function onDragHandleMouseUp(self, button)
	self:GetParent():StopMovingOrSizing()
end

local function onResize(self, width)
	rearrangeBars(self)
end

local function onDragStart(self)
	self:StartMoving()
end

local function onDragStop(self)
	self:StopMovingOrSizing()
	-- local p, f, r, x, y = self:GetPoint()
	-- self.pos_p = p
	-- self.pos_f = f
	-- self.pos_r = r
	-- self.pos_x = x
	-- self.pos_y = y
end

local function onControlEnter(self)
	GameTooltip:ClearLines()
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
	GameTooltip:AddLine(self.tooltipHeader)
	GameTooltip:AddLine(self.tooltipText, 1, 1, 1, 1)
	GameTooltip:Show()
end

local function onControlLeave()
	GameTooltip:Hide()
end

function lib:GetSpec(class, spec)
	if specs[class] and specs[class][spec] then
		return specs[class][spec]
	end
	return false
end

function lib:CreateAnchor(frameID, frameName)
	local f = CreateFrame("Frame", frameID, UIParent)
	f:SetWidth(200)
	f:SetHeight(16)

	local bg = f:CreateTexture(nil, "BACKGROUND")
	bg:SetTexture("Interface\\AddOns\\PvPTimer\\Textures\\waterline")
	bg:SetAllPoints(f)
	bg:SetBlendMode("BLEND")

	local headtext = f:CreateFontString(nil, "OVERLAY")
	headtext:SetFontObject(GameFontNormal)
	headtext:SetAllPoints(bg)
	headtext:SetJustifyH("CENTER")
	headtext:SetJustifyV("MIDDLE")
	headtext:SetShadowColor(0, 0, 0)
	headtext:SetShadowOffset(1, -1)

	local icon_class = f:CreateTexture(nil, "OVERLAY")
	icon_class:SetWidth(16)
	icon_class:SetHeight(16)
	icon_class:SetPoint("TOPLEFT", f, "TOPLEFT")

	local icon_spec = f:CreateTexture(nil, "OVERLAY")
	icon_spec:SetWidth(16)
	icon_spec:SetHeight(16)
	icon_spec:SetPoint("TOPRIGHT", f, "TOPRIGHT")
	
	f:SetScript("OnSizeChanged", onResize)
	f:SetScript("OnDragStart", onDragStart)
	f:SetScript("OnDragStop", onDragStop)
	
	f.ID = frameName
	f.running = {}
	f.columns = 3
	f.growup = false
	
	f.bg = bg
	f.headtext = headtext
	f.icon_class = icon_class
	f.icon_spec = icon_spec

	-- f.pos_p, f.pos_f, f.pos_r, f.pos_x, f.pos_y = f:GetPoint()

	return f
end

function lib:ClearAnchor(anchor)
	for bar in pairs(anchor.running) do
		bar:Stop()
		anchor.running[bar] = nil
	end
end

function lib:UnlockAnchor(anchor, state)
	if not anchor then return end

	if state then
		anchor:EnableMouse(true)
		anchor:SetMovable(true)
		anchor:SetResizable(true)
		anchor:RegisterForDrag("LeftButton")
	else
		anchor:EnableMouse(false)
		anchor:SetMovable(false)
		anchor:SetResizable(false)
		anchor:RegisterForDrag(nil)
	end
end

function lib:ResizeAnchor(anchor, width, height, scale)
	if not anchor then return end

	if width then anchor:SetWidth(width) end
	if height then anchor:SetHeight(height) end
	if scale then anchor:SetScale(scale) end
end

function lib:MoveAnchor(anchor, p, f, r, x, y)
	if not anchor then return end
	
	local old_p, old_f, old_r, old_x, old_y = anchor:GetPoint()
	if p == nil then p = old_p end
	if f == nil then f = old_f end
	if r == nil then r = old_r end
	if x == nil then x = old_x end
	if y == nil then y = old_y end

	anchor:ClearAllPoints()
	anchor:SetPoint(p, f, r, x, y)
end

function lib:SetHeader(anchor, name, class, spec, color)
	if not anchor then return end
	local ic = anchor.icon_class
	local is = anchor.icon_spec

	--set name
	if name then
		anchor.headtext:SetText(name)
	end

	--set bg color
	if color then
		anchor.bg:SetVertexColor(color.r, color.g, color.b, color.a)
	end

	--set class icon
	if class == "_HIDE_" then
		ic:Hide()
	elseif (class and CLASS_BUTTONS[class]) then
		if not ic:IsShown() then ic:Show() end
		ic:SetTexture("Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes")

		local left, right, top, bottom = unpack(CLASS_BUTTONS[class])
		left = left + (right - left) * 0.07
		right = right - (right - left) * 0.07
		top = top + (bottom - top) * 0.07
		bottom = bottom - (bottom - top) * 0.07
		ic:SetTexCoord(left, right, top, bottom)
	else
		if not ic:IsShown() then ic:Show() end
		ic:SetTexture(defaulticon)
		ic:SetTexCoord(0.05, 0.95, 0.05, 0.95)
	end

	--set spec icon
	if spec == "_HIDE_" then
		is:Hide()
	elseif (class and spec and specs[class] and specs[class][spec]) then
		if not is:IsShown() then is:Show() end
		
		is:SetTexture(specs[class][spec].icon)
	else
		if not is:IsShown() then is:Show() end
		
		is:SetTexture(defaulticon)
	end
	
	--rearrange text
	local justify = anchor.headtext:GetJustifyH()
	
	anchor.headtext:SetAllPoints(anchor.bg)
	if justify == "LEFT" then
		if ic:IsShown() then
			anchor.headtext:SetPoint("LEFT", anchor.icon_class, "RIGHT")
		end
	elseif justify == "RIGHT" then
		if is:IsShown() then
			anchor.headtext:SetPoint("RIGHT", anchor.icon_spec, "LEFT")
		end
	end
end

function lib:SetHeaderStyle(anchor, texture, font, fontsize, justify)
	if not anchor then return end

	--set bg texture
	if texture then
		if media:IsValid("statusbar", texture) then
			anchor.bg:SetTexture(media:Fetch("statusbar", texture))
		end
	end

	--set text font, size
	if font then
		if media:IsValid("font", font) then
			anchor.headtext:SetFont(media:Fetch("font", font), fontsize, "OUTLINE")
		end
	end
	
	--set text justify
	if justify == "LEFT" or justify == "CENTER" or justify == "RIGHT" then
		anchor.headtext:SetJustifyH(justify)
	end
end

function lib:SetBarStyle(anchor, texture, showname, showtime, namefont, namesize, timefont, timesize)
	if not anchor then return end
	
	anchor.texture = texture
	anchor.showname = showname
	anchor.showtime = showtime
	anchor.namefont = namefont
	anchor.namesize = namesize
	anchor.timefont = timefont
	anchor.timesize = timesize
end

function lib:UpdateSpell(anchor, spellID, duration, color)
	local bar = getBar(anchor, spellID)

	if not bar then
		bar = candy:New(defaulttexture, 64, 16)
		bar:Set("id", spellID)
		bar:Set("anchor", anchor)
		bar:SetIcon(select(3, GetSpellInfo(spellID)))
		candy.RegisterCallback(bar, "LibCandyBar_Stop", barstopped)
		anchor.running[bar] = true
		bar:SetDuration(duration)
		bar:Start()
	else
		bar:SetDuration(duration)
	end
	
	if media:IsValid("statusbar", anchor.texture) then
		bar:SetTexture(media:Fetch("statusbar", anchor.texture))
	end

	if anchor.showname then
		local name = GetSpellInfo(spellID)
		bar:SetLabel(name)
	else
		bar:SetLabel("")
	end
	
	bar:SetTimeVisibility(anchor.showtime)
	bar:SetScale(anchor:GetScale())
	bar:SetColor(color.r, color.g, color.b, color.a)

	rearrangeBars(anchor)
end
